# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
doge_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/doge'
    },
    {
        'url': 'https://dogecoin-api.flare.network/'
    },
    {
        'url': 'https://dogechain.rpc.thirdweb.com/'
    },
    {
        'url': 'https://rpc.dogechain.dog'
    }
]