# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
dogetest_nodes = [
    {
        'url': 'https://dogechain-testnet.rpc.thirdweb.com/' #!!!! these are ETHEREUM nodes, not DOGE nodes?
    },
    {
        'url': 'https://rpc-testnet.dogechain.dog' # Who's silly idea was that?
    },
]