# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
eth_nodes = [
    {
        'url': 'https://1rpc.io/eth'
    },
    {
        'url': 'https://ethereum-mainnet-rpc.allthatnode.com'
    },
    {
        'url': 'https://ethereum.publicnode.com'
    },
    {
        'url': 'https://rpc.ankr.com/eth'
    },
    {
        'url': 'https://eth-mainnet.public.blastapi.io'
    },
    {
        'url': 'https://ethereum.blinklabs.xyz/'
    },
    {
        'url': 'https://api.blockeden.xyz/eth/8UuXzatAZYDBJC6YZTKD'
    },
    {
        'url': 'https://ethereum.blockpi.network/v1/rpc/public'
    },
    {
        'url': 'https://eth-protect.rpc.blxrbdn.com'
    },
    {
        'url': 'https://cloudflare-eth.com/v1/mainnet'
    },
    {
        'url': 'https://eth.drpc.org/'
    },
    {
        'url': 'https://api.edennetwork.io/v1/rocket'
    },
    {
        'url': 'https://ethereum-bundler.etherspot.io/'
    },
    {
        'url': 'https://ethereum-api.flare.network/'
    },
    {
        'url': 'https://rpc.flashbots.net/'
    },
    {
        'url': 'https://rpc.eth.gateway.fm'
    },
    {
        'url': 'https://eth.meowrpc.com'
    },
    {
        'url': 'https://eth.llamarpc.com'
    },
    {
        'url': 'https://eth.merkle.io/'
    },
    {
        'url': 'https://rpc.mevblocker.io'
    },
    {
        'url': 'https://www.noderpc.xyz/rpc-mainnet/public'
    },
    {
        'url': 'https://lb.nodies.app/v1/5e9daed367d1454fab7c75f0ec8aceff'
    },
    {
        'url': 'https://endpoints.omniatech.io/v1/eth/mainnet/public'
    },
    {
        'url': 'https://eth.api.onfinality.io/public'
    },
    {
        'url': 'https://eth-pokt.nodies.app/'
    },
    {
        'url': 'https://rpc.propellerheads.xyz/eth'
    },
    {
        'url': 'https://mainnet.gateway.tenderly.co'
    },
    {
        'url': 'https://ethereum.rpc.thirdweb.com/'
    },
]
