# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
btc_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/btc',
    },
    {
        'url': 'https://bitcoin-mainnet-archive.allthatnode.com/'
    },
    {
        'url': 'https://bitcoin.drpc.org/'
    },
    {
        'url': 'https://bitcoin-api.flare.network'
    },
]

btc_esplora_nodes = [
    {
        'url': 'https://mempool.ninja'
    },
    {
        'url': 'https://www.lohkoketju.fi/'
    },
]