# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
ltc_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/ltc'
    },
    {
        'url': 'https://litecoin-api.flare.network/'
    },
]