# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
btctest_nodes = [
    {
        'url': 'https://bitcoin-testnet-archive.allthatnode.com/'
    },
    {
        'url': 'https://bitcoin-testnet.drpc.org/'
    },
]

btctest_esplora_nodes = [
    {
        'url': 'https://mempool.ninja/testnet/api'
    },
    {
        'url': 'https://www.lohkoketju.fi/testnet/api'
    },
]